import logging
from collections.abc import Iterable
from typing import List, Tuple, Union

import numpy as np
import torch
import torchvision.transforms.functional as F
from PIL import Image, ImageOps
from torch.nn.functional import pad
from torchvision.transforms.functional import InterpolationMode

from chariot_transforms.base import (
    Compose,
    IdentityTransform,
    ImageOnlyTransform,
    TransformBase,
    validate_rank_4,
)
from chariot_transforms.preprocessing.clahe import equalize_adapthist
from chariot_transforms.preprocessing.histogram_tools import linear_stretch


def get_height_width(img: Union[torch.Tensor, Image.Image]) -> Tuple[int, int]:
    if isinstance(img, torch.Tensor):
        if not len(img.shape) == 4:
            raise ValueError("img must be a rank 4 tensor")
        return tuple(img.shape[2:4])

    return (img.height, img.width)


class PreprocessingTransform(TransformBase):
    """These are the transforms which can be used as pre-processing transforms attribute of a model. It provides a common API
    for transforms which are used for reducing images to the size of a given NN and then transforming outputs (such as bounding boxes)
    back into the original image space.
    """

    # these work on batches
    @staticmethod
    def _validate_input(img):
        """here we allow PIL images as well as tensors for backwards compatibility"""
        if isinstance(img, Image.Image):
            return
        validate_rank_4(img)

    # convert_bbox and convert_coords can be public for pre-processing transforms since
    # they are used in model definitions to convert output but its best to keep them private for
    # augmentations
    def convert_bbox(self, bbox, img_h=None, img_w=None, invert=False):
        return self._convert_bbox(bbox, img_h, img_w, invert)

    def convert_coords(self, coords, img_h, img_w, invert=False):
        return self._convert_coords(coords, img_h, img_w, invert)


class HistogramLinearStretch(ImageOnlyTransform, PreprocessingTransform):
    # this works on a batch
    @staticmethod
    def _validate_input(img):
        validate_rank_4(img)

    def __init__(self, percentage: float) -> None:
        if (percentage <= 0) or (percentage >= 50):
            raise ValueError(
                "Linear Stretch must be a positive value smaller than 50."
            )
        if percentage > 45:
            logging.warning(
                f"Linear stretch of {percentage} is close to 50; this may result in occasional division by zero."
            )
            logging.warning(
                "Recommendation: use a linear stretch smaller than 5."
            )
        self.percentage = percentage

    def _convert_img_tensor(self, img: torch.Tensor) -> torch.Tensor:
        return torch.stack(
            [
                torch.tensor(linear_stretch(img[i].numpy(), self.percentage))
                for i in range(img.shape[0])
            ]
        )

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, HistogramLinearStretch)
            and self.percentage == other.percentage
        )


class CLAHE(PreprocessingTransform, ImageOnlyTransform):
    def __init__(self, nbins: int = 256) -> None:
        self.nbins = nbins

    def _convert_img_tensor(self, img: torch.Tensor) -> torch.Tensor:
        return torch.stack(
            [
                torch.tensor(
                    equalize_adapthist(
                        img[i].permute(1, 2, 0).numpy(), nbins=self.nbins
                    )  # need to put channels last because of `is_rgb_like` check in equalize_adapthisteq
                ).permute(2, 0, 1)
                for i in range(img.shape[0])
            ]
        )

    def __eq__(self, other: object) -> bool:
        return isinstance(other, CLAHE) and self.nbins == other.nbins


class TrivialPreprocessingTransform(PreprocessingTransform):
    """Base class for transforms that act trivially on annotations."""

    def _convert_bbox(self, bbox, img_h=None, img_w=None, invert=False):
        return bbox

    def _convert_mask(self, mask):
        return mask


class IdentityPreprocessingTransform(
    IdentityTransform, PreprocessingTransform
):
    """Transform used for models that do not need any image pre-processing.
    For all class methods the output is the input.
    """

    def __init__(self):
        pass

    def _convert_bbox(self, bbox, img_h=None, img_w=None, invert=False):
        return bbox

    def _convert_coords(self, coords, img_h, img_w, invert=False):
        return coords

    def set_magnitude(self, magnitude: float) -> None:
        return

    def __eq__(self, other):
        return isinstance(other, IdentityPreprocessingTransform)


class Resize(PreprocessingTransform):
    """Resize the input PIL Images to a given size or specified min/max size.

    This class method is able to resize torch tensors directly (on CPU or GPU)
    via its convert_img_pt method. Note that because of the way PIL and PyTorch
    resize, they may not always agree, i.e. if t is a Resize object then
    ToTensor()(t.convert_img(img)) and t.convert_img_pt(ToTensor()(img))
    will likely be different.
    """

    def __init__(
        self,
        size: Union[List[int], int] = None,
        min_size: int = None,
        max_size: int = None,
        interpolation: InterpolationMode = InterpolationMode.BILINEAR,
    ) -> None:
        """
        Parameters
        ----------
        size : sequence or int
            Desired output size. If size is a sequence like
            (h, w), output size will be matched to this. If size is an int,
            smaller edge of the image will be matched to this number.
            i.e, if height > width, then image will be rescaled to
            (size * height / width, size)
        min_size : int
            Desired minimum size of single dimension (height or width) of an image.
        max_size : int
            Desired maximum size of single dimension (height or width) of an image.
        interpolation : int
            Desired interpolation. Default is ``InterpolationMode.BILINEAR``


        Note: Exactly one of size or (min_size and max_size) must be set.
        If size is set, every image will be resized to match the specified size.
        If (instead) min_size and max_size are set, each input image will be resized so that
        either the resized_min_size matches the specified min_size or the
        resized_max_size matches the specified max_size, whichever results in the resized image
        having smaller area.  It is not guaranteed that the resized image will satisfy both
        (weak) constraints: i.e. that resized_max_size <= max_size and resized_min_size <= min_size.
        """

        assert bool(min_size) == bool(
            max_size
        ), "If you set one of min_size or max_size, you must set both (and be positive integers)."
        assert bool(size) != bool(
            min_size
        ), "You must set either size or (min_size and max_size)."
        if size is not None:
            assert isinstance(size, int) or (
                isinstance(size, Iterable) and len(size) == 2
            )
        else:
            assert min_size > 0, "min_size must be a positive integer"
            assert max_size >= min_size, "max_size must be at least min_size"

        self.size = size
        self.interpolation = interpolation
        self.min_size = min_size
        self.max_size = max_size

    def get_new_size(self, img_shape: Tuple[int, int]) -> Tuple[int, int]:
        """
        Parameters
        ----------
        img_shape
            tuple of the form (height, width)

        Returns
        -------
        Tuple[int, int]
            the height and width of the resized image
        """
        if self.size is not None:
            if isinstance(self.size, int):
                return (
                    (self.size * img_shape[0] / img_shape[1], self.size)
                    if img_shape[0] > img_shape[1]
                    else (self.size, self.size * img_shape[1] / img_shape[0])
                )
            return self.size
        else:
            # both self.min_size and self.max_size are set...
            # determine current min/max lengths.
            current_min_side_length = min(img_shape)
            current_max_side_length = max(img_shape)

            # compute possible resizing ratios (scales)
            tight_min_ratio = self.min_size / current_min_side_length
            tight_max_ratio = self.max_size / current_max_side_length

            # best ratio is the smaller of the two
            ratio = min(tight_min_ratio, tight_max_ratio)

            return (int(img_shape[0] * ratio), int(img_shape[1] * ratio))

    def _calc_resize(self, img_shape):
        new_size = self.get_new_size(img_shape)
        size = (np.array(new_size, dtype=int)).astype(float)
        return size[0] / img_shape[0], size[1] / img_shape[1]

    def _convert_img_tensor(self, img: Union[torch.Tensor, Image.Image]):
        # set resize for converting bounding boxes
        self.orig_size = get_height_width(img)
        size = (
            self.size
            if self.size is not None
            else self.get_new_size(self.orig_size)
        )
        return F.resize(img, size, self.interpolation, antialias=True)

    def _convert_mask(self, mask, invert=False):
        if invert:
            raise ValueError("inversion not supported")
        return super()._convert_mask(mask)

    def _convert_bbox(self, bbox, img_h=None, img_w=None, invert=False):
        """Converts a bounding box. If invert is False then transforms the
        bounding box from an img_h x img_w image to the bounding box for the
        resized self.new_h x self.new_w image. If invert is True then it will do the
        conversion in the opposite direction.

        Parameters
        ----------
        bbox : list
            of the form [ymin, xmin, ymax, xmax]
        img_h : int
            if None then will use self.img_h
        img_w : int
            if None then will use self.img_w
        invert : bool
            whether or not to return the inverse operation

        Returns
        -------
        list
            list of integers of the transformed bounding box.
        """

        if not hasattr(self, "orig_size") and not (img_h and img_w):
            raise Exception(
                "Must call convert_img before convert_bbox or pass img_h and img_w"
            )
        orig_size = (img_h, img_w) if (img_h and img_w) else self.orig_size
        resize = self._calc_resize(orig_size)
        exp = -1 if invert else 1
        return [
            resize[0] ** exp,
            resize[1] ** exp,
            resize[0] ** exp,
            resize[1] ** exp,
        ] * np.array(bbox)

    def _convert_coords(self, coords, img_h, img_w, invert=False):
        """Converts coordinates. If invert is False then transforms the
        coordinates from an img_h x img_w image to the coordinates for the
        resized new_h x new_w image. If invert is True then it will do the
        conversion in the opposite direction.

        Parameters
        ----------
        coords : np.ndarray
            of shape N(x...)x2 of the form [[y_0, x_0], [y_1, x_1], ...]
        img_h : int
        img_w : int
        invert : bool
            whether or not to return the inverse operation

        Returns
        -------
        np.ndarray
            array of the transformed coordinates.
        """
        exp = -1 if invert else 1
        resize = np.array(self._calc_resize((img_h, img_w))) ** exp
        new_coords = np.array(coords) * resize
        return new_coords

    def __repr__(self):
        if self.size is not None:
            return (
                self.__class__.__name__
                + "(size={0}, interpolation={1})".format(
                    self.size, self.interpolation
                )
            )
        else:
            return (
                self.__class__.__name__
                + "(min_size={0}, max_size={1}, interpolation={2})".format(
                    self.min_size, self.max_size, self.interpolation
                )
            )

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.size == other.size
            and self.interpolation == other.interpolation
            and self.min_size == other.min_size
            and self.max_size == other.max_size
        )


class ResizePreserveAspect(PreprocessingTransform):
    """Class for resizing an image and bounding boxes to (new_h, new_w)
    by preserving aspect ratio and padding as necessary.

    This class method is able to resize torch tensors directly (on CPU or GPU)
    via its convert_img_pt method. Note that because of the way PIL and PyTorch
    resize, they may not always agree, i.e. if t is a ResizePreserveAspect
    object then ToTensor()(t.convert_img(img)) and
    t.convert_img_pt(ToTensor()(img)) will likely be different.
    """

    # this works on a batch
    @staticmethod
    def _validate_input(img):
        validate_rank_4(img)

    def __init__(
        self,
        new_h: int,
        new_w: int,
        fill: Union[int, Tuple[int, int, int]] = 0,
    ):
        """
        Parameters
        ----------
        new_h : int
            height to resize to
        new_w : int
            width to resize to
        fill : int or Tuple[int, int, int]
            the fill color to use for the padding
        """
        self.new_h = new_h
        self.new_w = new_w
        self.fill = fill

    def _convert_bbox(self, bbox, img_h=None, img_w=None, invert=False):
        """Converts a bounding box. If invert is False then transforms the
        bounding box from an img_h x img_w image to the bounding box for the
        resized self.new_h x self.new_w image. If invert is True then it will do the
        conversion in the opposite direction.

        Parameters
        ----------
        bbox : list
            of the form [ymin, xmin, ymax, xmax]
        img_h : int
            if None then will use self.img_h
        img_w : int
            if None then will use self.img_w
        invert : bool
            whether or not to return the inverse operation

        Returns
        -------
        list
            list of integers of the transformed bounding box.
        """

        img_w = img_w or self.img_w
        img_h = img_h or self.img_h

        new_w = self.new_w
        new_h = self.new_h

        if img_w / img_h <= new_w / new_h:
            # padding was added to increase width
            act_w = int(img_w / img_h * new_h)  # image width without padding
            x_offset = (new_w - act_w) // 2
            x_scale = act_w / img_w

            y_offset = 0
            y_scale = new_h / img_h
        else:
            # padding was added to increase height
            act_h = int(img_h / img_w * new_w)  # image height without padding
            y_offset = (new_h - act_h) // 2
            y_scale = act_h / img_h

            x_offset = 0
            x_scale = new_w / img_w

        if invert:
            return [
                (bbox[0] - y_offset) / y_scale,
                (bbox[1] - x_offset) / x_scale,
                (bbox[2] - y_offset) / y_scale,
                (bbox[3] - x_offset) / x_scale,
            ]
        return [
            y_scale * bbox[0] + y_offset,
            x_scale * bbox[1] + x_offset,
            y_scale * bbox[2] + y_offset,
            x_scale * bbox[3] + x_offset,
        ]

    def _convert_coords(self, coords, img_h, img_w, invert=False):
        """Converts coordinates. If invert is False then transforms the
        coordinates from an img_h x img_w image to the coordinates for the
        resized new_h x new_w image. If invert is True then it will do the
        conversion in the opposite direction.

        Parameters
        ----------
        coords : np.ndarray
            of shape N(x...)x2 of the form [[y_0, x_0], [y_1, x_1], ...]
        img_h : int
        img_w : int
        invert : bool
            whether or not to return the inverse operation

        Returns
        -------
        np.ndarray
            array of the transformed coordinates.
        """

        new_w = self.new_w
        new_h = self.new_h

        if img_w / img_h <= new_w / new_h:
            # padding was added to increase width
            act_w = int(img_w / img_h * new_h)  # image width without padding
            x_offset = (new_w - act_w) // 2
            x_scale = act_w / img_w

            y_offset = 0
            y_scale = new_h / img_h
        else:
            # padding was added to increase height
            act_h = int(img_h / img_w * new_w)  # image height without padding
            y_offset = (new_h - act_h) // 2
            y_scale = act_h / img_h

            x_offset = 0
            x_scale = new_w / img_w

        new_coords = np.zeros(coords.shape, dtype=coords.dtype)
        if invert:
            new_coords[..., 0] = (coords[..., 0] - y_offset) / y_scale
            new_coords[..., 1] = (coords[..., 1] - x_offset) / x_scale
        else:
            new_coords[..., 0] = coords[..., 0] * y_scale + y_offset
            new_coords[..., 1] = coords[..., 1] * x_scale + x_offset

        return new_coords

    def _convert_pil(self, img: Image.Image) -> Image.Image:
        w, h = img.size
        self.img_w, self.img_h = w, h

        if w / h <= self.new_w / self.new_h:
            act_w = int(w / h * self.new_h)
            new_img = img.resize((act_w, self.new_h), Image.BILINEAR)
            border = self.new_w - act_w
            # add border // 2 on the left and border - border // 2 on the right
            new_img = ImageOps.expand(
                new_img, (border // 2, 0, border - border // 2, 0), self.fill
            )
        else:
            act_h = int(h / w * self.new_w)
            new_img = img.resize((self.new_w, act_h), Image.BILINEAR)
            border = self.new_h - act_h
            # add border // 2 on the top and border - border // 2 on the bottom
            new_img = ImageOps.expand(
                new_img, (0, border // 2, 0, border - border // 2), self.fill
            )
        return new_img

    def _convert_img_tensor(self, img: Union[torch.Tensor, Image.Image]):
        if isinstance(img, Image.Image):
            return self._convert_pil(img)

        h, w = get_height_width(img)
        self.img_h, self.img_w = h, w

        if w / h <= self.new_w / self.new_h:
            act_w = int(w / h * self.new_h)
            new_img = F.resize(
                img,
                (self.new_h, act_w),
                interpolation=InterpolationMode.BILINEAR,
                antialias=True,
            )
            border = self.new_w - act_w
            # add border // 2 on the left and border - border // 2 on the right
            new_img = pad(
                new_img, (border // 2, border - border // 2), value=self.fill
            )
        else:
            act_h = int(h / w * self.new_w)
            new_img = F.resize(
                img,
                (act_h, self.new_w),
                interpolation=InterpolationMode.BILINEAR,
                antialias=True,
            )
            border = self.new_h - act_h
            # add border // 2 on the top and border - border // 2 on the bottom
            new_img = pad(
                new_img,
                (0, 0, border // 2, border - border // 2),
                value=self.fill,
            )
        return new_img

    def _convert_mask(self, mask, invert=False):
        if invert:
            raise ValueError("Inversion not supported")
        return super()._convert_mask(mask)

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.new_h == other.new_h
            and self.new_w == other.new_w
        )


class PreprocessingCompose(Compose):
    def __init__(self, transforms: List[PreprocessingTransform]):
        super().__init__(transforms)
        self.resize_component = None
        for t in transforms:
            if isinstance(t, RESIZE_TRANSFORMS):
                if self.resize_component is not None:
                    raise ValueError(
                        "PreprocessingCompose cannot contain more than one resize transform"
                    )
                self.resize_component = t


class PixelTransform(ImageOnlyTransform, PreprocessingTransform):
    def __init__(
        self,
        mean: List[float] = None,
        std: List[float] = None,
        bgr: bool = False,
    ):
        """Does pixelwise normalization

        Parameters
        ----------
        mean
            RGB-mean to use for normalization
        std
            RGB-standard deviation to use for normalization
        bgr
            if image should be converted into BGR or not
        """
        # for serialization purposes we set mean and std here despite defining
        # mean_tensor and std_tensor later
        self.mean = mean
        self.std = std
        self.bgr = bgr

        self.mean_tensor = (
            torch.tensor(mean).reshape(1, len(mean), 1, 1)
            if mean is not None
            else None
        )
        self.std_tensor = (
            torch.tensor(std).reshape(1, len(std), 1, 1)
            if std is not None
            else None
        )

    def _convert_img_tensor(self, x):
        if self.mean_tensor is not None and self.std_tensor is not None:
            x = (x - self.mean_tensor) / self.std_tensor

        if self.bgr:
            return x[:, [2, 1, 0], :, :]

        return x

    def __eq__(self, other):
        if not isinstance(other, PixelTransform):
            return False
        return (
            (self.mean == other.mean)
            and (self.std == other.std)
            and (self.bgr == other.bgr)
        )


RESIZE_TRANSFORMS = (Resize, ResizePreserveAspect)



import random

import numpy as np
import torch
from PIL import Image, ImageDraw
from torchvision.transforms.functional import to_tensor

from chariot_transforms import bbox_ops
from chariot_transforms.augmentations.transforms import (
    AutoContrast,
    Brightness,
    CenterCrop,
    ColorJitter,
    Contrast,
    Cutout,
    Equalize,
    Gamma,
    Grayscale,
    HorizontalShear,
    Hue,
    Invert,
    Pad,
    Posterize,
    RandAugment,
    RandomAffine,
    RandomCrop,
    RandomGrayscale,
    RandomHorizontalFlip,
    RandomPerspective,
    RandomResize,
    RandomRotation,
    RandomVerticalFlip,
    Resize,
    ResizePreserveAspect,
    Rotation,
    Saturation,
    Sharpness,
    Solarize,
    TranslateX,
    TranslateY,
)

bbox_ops.DROP_BBOX_AREA_RATIO_THRES = 0.0

random.seed(7)


def random_bbox(img):
    ymin = random.randint(0, img.height - 40)
    ymax = random.randint(ymin + 20, img.height)
    xmin = random.randint(0, img.width - 40)
    xmax = random.randint(xmin + 20, img.width)

    return [ymin, xmin, ymax, xmax]


eps = 1e-6


def check_image(img):
    assert len(img.shape) == 4
    assert img.min().item() >= 0 - eps
    assert img.max().item() <= 1 + eps


def _test_bbox_conversion(transform, atol: float = 3.5):
    """
    Test that bounding boxes get transformed properly with an image.
    It does this by doing the following 100 times:

    1. Create a PIL image with a random rectangle in it.
    2. Transform the image and bounding box of the rectangle
    3. Check that the resulting rectangle in the image is
    circumscribed by the transformed bounding box
    """

    for _ in range(100):
        img = Image.new("RGB", (200, 100))
        bbox = random_bbox(img)
        img_draw = ImageDraw.Draw(img)
        img_draw.rectangle([bbox[1], bbox[0], bbox[3], bbox[2]], fill="red")

        img_tensor = to_tensor(img).unsqueeze(0)
        trans_img, trans_bbox_dict = transform(
            img_tensor, bbox_dict={"bboxes": [bbox], "classes": ["class"]}
        )

        # find pixels of the rectangle in the image
        _, ys, xs = torch.where(trans_img[0] > 20 / 255)

        if len(ys) == 0 or ys.max() - ys.min() < 1 or xs.max() - xs.min() < 1:
            # if no pixels (e.g. if rectangle got cropped out) check that
            # no bounding boxes come back
            assert len(trans_bbox_dict["bboxes"]) == 0
        elif ys.max() - ys.min() == 1 or xs.max() - xs.min() == 1:
            # a sliver
            assert len(trans_bbox_dict["bboxes"]) <= 1
        else:
            assert len(trans_bbox_dict["bboxes"]) != 0
            trans_bbox = trans_bbox_dict["bboxes"][0]
            np.testing.assert_allclose(
                [ys.min(), xs.min(), ys.max(), xs.max()],
                trans_bbox,
                rtol=1e-3,
                atol=atol,
            )


def test_auto_contrast():
    _test_bbox_conversion(AutoContrast())


def test_brightness():
    _test_bbox_conversion(Brightness(0.7))


def test_saturation():
    _test_bbox_conversion(Saturation(0.7))


def test_center_crop():
    _test_bbox_conversion(CenterCrop(size=(12, 20)))


def test_color_jitter():
    _test_bbox_conversion(ColorJitter(0.2, 0.2, 0.2, 0.2))


def test_contrast():
    _test_bbox_conversion(Contrast(1.2))


def test_gamma():
    _test_bbox_conversion(Gamma(1.2))


def test_hue():
    _test_bbox_conversion(Hue(0.2))


def test_cutout():
    _test_bbox_conversion(Cutout(16))


def test_equalize():
    _test_bbox_conversion(Equalize())


def test_grayscale():
    _test_bbox_conversion(Grayscale(num_output_channels=3))


def test_horizontal_shear():
    _test_bbox_conversion(HorizontalShear(0.1))


def test_invert():
    img = to_tensor(Image.new("RGB", (200, 100))).unsqueeze(0)
    new_img = Invert()(img)
    assert new_img.max() <= 1
    assert new_img.min() >= 0
    assert new_img.shape == img.shape
    torch.testing.assert_close(1 - img, new_img)


def test_pad():
    _test_bbox_conversion(Pad(padding=(16, 14, 32, 6)))


def test_posterize():
    _test_bbox_conversion(Posterize(3))


def test_random_crop():
    _test_bbox_conversion(RandomCrop((100, 50)))


def test_random_horizontal_flip():
    _test_bbox_conversion(RandomHorizontalFlip())


def test_horizontal_flip():
    _test_bbox_conversion(RandomHorizontalFlip(1))


def test_random_resize():
    _test_bbox_conversion(RandomResize(scale_range_high=2, p=1.0), atol=3)


def test_random_vertical_flip():
    _test_bbox_conversion(RandomVerticalFlip())


def test_random_affine():
    _test_bbox_conversion(
        RandomAffine(
            degrees=0, translate=(0.1, 0.1), scale=(0.9, 1.2), shear=(-10, 10)
        ),
        atol=10,
    )


def test_rand_augment_m_specified():
    _test_bbox_conversion(
        RandAugment(
            n=2,
            m=3,
            augs_to_remove=[
                "Invert",
                "Solarize",
                "Brightness",
                "Contrast",
                "HorizontalShear",
                "VerticalShear",
            ],
        ),
        atol=8.5,
    )


def test_rand_augment_m_not_specified():
    _test_bbox_conversion(
        RandAugment(
            n=3,
            augs_to_remove=[
                "Invert",
                "Solarize",
                "Brightness",
                "Contrast",
                "Rotation",
                "HorizontalShear",
                "VerticalShear",
                "TranslateY",
                "TranslateX",
            ],
        )
    )


def test_random_grayscale():
    _test_bbox_conversion(RandomGrayscale())


def test_random_perspective():
    _test_bbox_conversion(RandomPerspective(0.1), atol=11)


def test_random_rotation():
    _test_bbox_conversion(RandomRotation(10))


def test_resize():
    _test_bbox_conversion(Resize((32, 32)))


def test_resize_single_int():
    img = to_tensor(Image.new("RGB", (200, 100))).unsqueeze(0)
    new_img = Resize(32)(img)
    assert new_img.shape == (1, 3, 32, 32 * 200 / 100)
    img = to_tensor(Image.new("RGB", (100, 200))).unsqueeze(0)
    new_img = Resize(32)(img)
    assert new_img.shape == (1, 3, 32 * 200 / 100, 32)


def test_resize_preserve_aspect():
    _test_bbox_conversion(ResizePreserveAspect(new_h=32, new_w=32))


def test_rotation():
    _test_bbox_conversion(Rotation(10))


def test_sharpness():
    _test_bbox_conversion(Sharpness(0.5))


def test_solarize():
    img = Image.new("RGB", (200, 100))
    img_tensor = to_tensor(img).unsqueeze(0)

    check_image(Solarize(1)(img_tensor))
    check_image(Solarize(0)(img_tensor))
    check_image(Solarize(0.5)(img_tensor))

    # with threshold 1 the transform shouldn't do anything
    _test_bbox_conversion(Solarize(1))


def test_translate_x():
    _test_bbox_conversion(TranslateX(0.4))


def test_translate_y():
    _test_bbox_conversion(TranslateY(0.4))



from abc import abstractmethod
from typing import List

import torch
import torchvision.transforms.functional as F
from PIL import Image


def validate_rank_4(img):
    if isinstance(img, torch.Tensor) and len(img.shape) != 4:
        raise ValueError("img must be a rank 4 tensor")


class TransformBase:
    """Base class for both pre-processing and augmentation transforms"""

    @staticmethod
    def _validate_input(img):
        validate_rank_4(img)
        if img.shape[0] != 1:
            raise ValueError(
                "Transforms currently only support batches of size 1."
            )

    def __call__(
        self,
        img: torch.Tensor,
        mask: Image.Image = None,
        bbox_dict: dict = None,
    ):
        """
        Parameters
        ----------
        img
            should be a rank 4 image tensor. Currently we only support batches of size 1
            (i.e. img.shape[0] should be 1)
        mask : Image.Image
            segmentation mask
        bbox_dict : dict
            dict with keys "bboxes" and "classes"

        Returns
        -------
        torch.Tensor or tuple
            If only img is passed then the return is the transformed image. Otherwise it is
            a tuple with the first item the transformed image and the other items
            the transformed annotations.
        """
        self._validate_input(img)
        ret = [self._convert_img_tensor(img)]
        if mask is not None:
            ret.append(self._convert_mask(mask))
        if bbox_dict is not None:
            new_bbox_dict = {"bboxes": [], "classes": []}
            for bbox, label in zip(bbox_dict["bboxes"], bbox_dict["classes"]):
                new_bbox = self._convert_bbox(bbox)
                if new_bbox is not None:
                    new_bbox_dict["bboxes"].append(new_bbox)
                    new_bbox_dict["classes"].append(label)

            ret.append(new_bbox_dict)

        return tuple(ret) if len(ret) > 1 else ret[0]

    @abstractmethod
    def _convert_img_tensor(self, img: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError

    @abstractmethod
    def _convert_bbox(self, bbox, img_h=None, img_w=None, invert=False):
        raise NotImplementedError

    @abstractmethod
    def _convert_coords(self, coords, img_h, img_w, invert=False):
        raise NotImplementedError

    def _convert_mask(self, mask: Image.Image) -> Image.Image:
        """
        Parameters
        ----------
        mask
            mask should be an 8-bit, single channel PIL image
        """
        return F.to_pil_image(
            self._convert_img_tensor(F.to_tensor(mask).unsqueeze(0))[0]
        )


class ImageOnlyTransform(TransformBase):
    """Base class for transforms that do not change bounding boxes or masks, such as
    color jitters, histogram transforms, etc.
    """

    def _convert_bbox(self, bbox, img_h=None, img_w=None, invert=False):
        return bbox

    def _convert_coords(self, coords, img_h, img_w, invert=False):
        return coords

    def _convert_mask(self, mask):
        return mask


class IdentityTransform(ImageOnlyTransform):
    def __init__(self):
        pass

    def _convert_bbox(self, bbox, img_h=None, img_w=None, invert=False):
        return bbox

    def _convert_coords(self, coords, img_h, img_w, invert=False):
        return coords

    def set_magnitude(self, magnitude: float) -> None:
        return

    def _convert_img_tensor(self, img: torch.Tensor) -> torch.Tensor:
        return img

    def __eq__(self, other):
        return isinstance(other, IdentityTransform)


class Compose(TransformBase):
    """Composes several transforms together."""

    __serialization_attributes__ = ["transforms"]

    def __init__(self, transforms: List[TransformBase]):
        """
        Parameters
        ----------
        transforms : list
            list of transforms to compose
        """
        if len(transforms) == 0:
            raise ValueError("transforms list should not be empty")

        self.transforms = []
        for t in transforms:
            if isinstance(t, Compose):
                self.transforms.extend(t.transforms)
            else:
                self.transforms.append(t)

    def _convert_img_tensor(self, img):
        for t in self.transforms:
            img = t._convert_img_tensor(img)
        return img

    def _convert_mask(self, mask):
        for t in self.transforms:
            mask = t._convert_mask(mask)
        return mask

    def _convert_bbox(self, bbox):
        for t in self.transforms:
            bbox = t._convert_bbox(bbox)
            if bbox is None:
                return None
        return bbox

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
        format_string += "\n)"
        return format_string

    def __eq__(self, other: object) -> bool:
        # ignore identity transforms when comparing
        self_transforms = [
            t for t in self.transforms if not isinstance(t, IdentityTransform)
        ]

        if not isinstance(other, Compose):
            # consider a Compose with a single element to be the same
            if isinstance(other, TransformBase) and len(self_transforms) == 1:
                return self.transforms[0] == other
            if (
                isinstance(other, TransformBase)
                and isinstance(other, IdentityTransform)
                and len(self_transforms) == 0
            ):
                return True
            return False

        other_transforms = [
            t for t in other.transforms if not isinstance(t, IdentityTransform)
        ]
        if len(self_transforms) != len(other_transforms):
            return False
        for t1, t2 in zip(self_transforms, other_transforms):
            if t1 != t2:
                return False
        return True



from .transforms import (
    CLAHE,
    HistogramLinearStretch,
    IdentityPreprocessingTransform,
    PixelTransform,
    PreprocessingCompose,
    PreprocessingTransform,
    Resize,
    ResizePreserveAspect,
    TrivialPreprocessingTransform,
)

__all__ = [
    "CLAHE",
    "HistogramLinearStretch",
    "IdentityPreprocessingTransform",
    "PixelTransform",
    "PreprocessingTransform",
    "TrivialPreprocessingTransform",
    "Resize",
    "ResizePreserveAspect",
    "PreprocessingCompose",
]



import pytest

from chariot_transforms.augmentations import Pad
from chariot_transforms.base import Compose
from chariot_transforms.preprocessing import (
    IdentityPreprocessingTransform,
    Resize,
)


def test_compose_equality():
    t1 = IdentityPreprocessingTransform()
    t2 = Resize(224)
    t3 = Pad(10)

    assert Compose([t1, t2, t3]) == Compose([t2, t3])
    assert Compose([t2]) == t2
    assert Compose([t2, Compose([t1, t3])]) == Compose([t2, t3])

    assert Compose([t1, t3]) != Compose([t1])
    assert Compose([t1]) != t2


def test_compose_validation():
    with pytest.raises(ValueError) as exc_info:
        Compose([])
    assert "transforms list should not be empty" in str(exc_info.value)



from typing import Optional, Union

import numpy as np


def area(bbox):
    return (bbox[2] - bbox[0]) * (bbox[3] - bbox[1])


def intersection(bbox1, bbox2):
    lt = np.max([bbox1[:2], bbox2[:2]], axis=0)
    rb = np.min([bbox1[2:], bbox2[2:]], axis=0)

    wh = np.clip(rb - lt, 0, None)
    return wh[0] * wh[1]


def iou(bbox1, bbox2):
    inter = intersection(bbox1, bbox2)
    return inter / (area(bbox1) + area(bbox2) - inter)


# transformed bounding boxes that have a ratio of new area to old area
# less than this threshold will be dropped.
DROP_BBOX_AREA_RATIO_THRES = 0.1


def clean_bbox(
    bbox: list,
    img_height: int = np.infty,
    img_width: int = np.infty,
    og_bbox: Optional[list] = None,
) -> Union[list, None]:
    """Cleans a transformed bounding box by clipping the x-coordinate
    components to be between 0 and img_width and clipping the y-coordinate
    components to be between 0 and img_height. It also checks if the bounding
    box has length or width less than or equal to 1, in which case it will
    return None. If `og_bbox` is passed then None will be returned if the ratio
    of the area of `bbox` to the area of `og_bbox` is less than DROP_BBOX_AREA_RATIO_THRES

    Parameters
    ----------
    bbox
        bounding box, of form [ymin, xmin, ymax, xmax], to clean
    img_height
        height of image the bounding box corresponds to
    img_width
        width of image the bounding box corresponds to
    og_bbox
        (optional) original bounding box that `bbox` was gotten from via a
        transformation

    Returns
    -------
    cleaned bounding box, as a list of length four, or None if the bounding
    box is not sufficiently large.
    """
    bbox = np.clip(
        bbox,
        [0] * 4,
        [img_height - 1, img_width - 1, img_height - 1, img_width - 1],
    )

    if bbox[2] - bbox[0] <= 1 or bbox[3] - bbox[1] <= 1:
        return None

    if og_bbox is not None:
        if area(bbox) / area(og_bbox) < DROP_BBOX_AREA_RATIO_THRES:
            return None

    return bbox


def rotate_bounding_box(angle: float, center: list, bbox: list) -> list:
    """
    Parameters
    ----------
    angle
        angle, in degrees, to rotate
    center
        (x, y) point around which to rotate
    bbox
        bounding box, of form [ymin, xmin, ymax, xmax], to rotate

    Returns
    -------
    rotated bounding box, as a list of length four.
    """
    center = np.array(center)
    angle_rad = np.pi * angle / 180

    rot_matrix = np.array(
        [
            [np.cos(angle_rad), np.sin(angle_rad)],
            [-np.sin(angle_rad), np.cos(angle_rad)],
        ]
    )

    points = np.array(
        [
            [bbox[1], bbox[1], bbox[3], bbox[3]],
            [bbox[0], bbox[2], bbox[0], bbox[2]],
        ]
    )
    new_points = rot_matrix @ (
        points - center.reshape(2, -1)
    ) + center.reshape(2, -1)

    minpts = new_points.min(1)
    maxpts = new_points.max(1)

    return [minpts[1], minpts[0], maxpts[1], maxpts[0]]


def apply_affine_matrix_to_bbox(matrix: np.ndarray, bbox: list) -> list:
    """Transforms a bounding box via an affine matrix

    Parameters
    ----------
    matrix
        3x3 affine matrix acting on the plane
    bbox
        list of length four of the form [y_min, x_min, y_max, x_max]

    Returns
    -------
    transformed bounding box
    """
    points = np.array(
        [
            [bbox[1], bbox[1], bbox[3], bbox[3]],
            [bbox[0], bbox[2], bbox[0], bbox[2]],
            [1.0, 1.0, 1.0, 1.0],
        ]
    )

    new_points = (matrix @ points)[:2]

    minpts = new_points.min(1)
    maxpts = new_points.max(1)

    return [minpts[1], minpts[0], maxpts[1], maxpts[0]]



import random

import numpy as np
import pytest
import torch

from chariot_transforms.augmentations.transforms import RandAugment


@pytest.fixture
def img():
    return torch.rand(1, 3, 10, 10)


def test_augments(img):
    """Test that rand augment runs and changes an image and keeps its size"""
    rand_augment = RandAugment(10, 2)

    trans_img = rand_augment(img)
    assert (img != trans_img).all()
    assert img.shape == trans_img.shape


def test_augments_no_m(img):
    """Test that rand augment with random magnitude runs and changes an
    image and keeps its size
    """
    rand_augment = RandAugment(10)
    trans_img = rand_augment(img)
    assert not (img == trans_img).all()
    assert img.shape == trans_img.shape


def test_specified_augmentation_identity(img):
    """Test that augmentations can be specified"""
    rand_augment = RandAugment(10, augs_to_use=["Identity"])
    trans_img = rand_augment(img)
    assert (img == trans_img).all()
    assert img.shape == trans_img.shape


def test_specified_augmentations(img):
    """Test that augmentations can be specified"""
    random.seed(42)
    rand_augment = RandAugment(10, 3, augs_to_use=["Equalize", "Invert"])
    trans_img = rand_augment(img)
    for a in rand_augment.augmentations:
        assert a[0] in ["Equalize", "Invert"]

    assert not (img == trans_img).all()
    assert img.shape == trans_img.shape


def test_restricted_augmentations(img):
    """Test that augmentations can be removed"""
    rand_augment = RandAugment(10, 3, augs_to_remove=["Equalize", "Identity"])
    trans_img = rand_augment(img)
    for a in rand_augment.augmentations:
        assert a[0] not in ["Color", "Identity"]

    assert (
        len(RandAugment.augmentations) == len(rand_augment.augmentations) + 2
    )

    assert not (np.array(img) == np.array(trans_img)).all()
    assert img.shape == trans_img.shape


def test_mag_too_big():
    """Test that an error is raised if a magnitude larger than 11 is passed"""
    with pytest.raises(ValueError) as exc_info:
        RandAugment(10, 11)
    assert "m must be between 0 and 10" in str(exc_info.value)


def test_mag_too_small():
    """Test that an error is raised if a magnitude larger than 0 is passed"""
    with pytest.raises(ValueError) as exc_info:
        RandAugment(10, -1)
    assert "m must be between 0 and 10" in str(exc_info.value)
